/* NAME:
 *	runChild - run a command
 *
 * SYNOPSIS:
 *	void setChildAv(char * "cmd", int "nargs")
 *	void runChild(char * "cmd", int "nargs", char **"args", int "async")
 *	void reapChildren()
 *
 * DESCRIPTION:
 *	Runs "cmd" with "args" appended if not null.
 *	If "async" is not set, we wait for the child to terminate
 *	otherwise we reap any dead children but otherwise don't worry
 *	too much.
 *	
 *	"Cmd" may be a comma separated list of words each of which
 *	will get its own arg in the resulting command.
 *	If "cmd" is null when runChild is called, the previous av will
 *	be used.
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 *	
 */
/*
 *	@(#)Copyright (c) 1993 QUICK.COM.AU. All rights reserved.
 *	
 *	This file is provided in the hope that it will
 *	be of use.  There is absolutely NO WARRANTY.
 *	Permission to use, modify and distribute this source code 
 *	is granted subject to the following conditions.
 *	1/ that the above copyright notice and this notice 
 *	are preserved in all copies and that due credit be given 
 *	to the author.	
 *	2/ that any changes to this code are clearly commented 
 *	as such so that the author does not get blamed for bugs 
 *	other than his own.
 *	3/ If there is an accompanying License file its provisions
 *	must also be met.
 *	
 *	Please send copies of changes and bug-fixes to:
 *	sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: runChild.c,v 1.3 1999/07/15 07:14:57 sjg Exp $";
#endif
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/wait.h>

static char **Av = 0;
static int Ac = 0;
static int Nargs = 0;

int
setChildAv(cmd, nargs)
	char *cmd;
	int nargs;
{
	static int nav = 0;
	char *cp;
	int i;

	if (!cmd)
		return;
	Ac = 0;
#ifdef MAIN
	printf("setChildAv('%s')\n", cmd);
#endif
	
	/*
	 * work out how many ',' separated words in "cmd".
	 */
	for (cp = cmd, i = nargs + 1; 
	     cp;
	     cp = strchr(cp+1, ','))
		++i;
	if (nav == 0) {
		Av = (char **) malloc(i * sizeof (char *));
		nav = i;
	} else if (nav < i) {
		Av = (char **) realloc(Av, i * sizeof (char *));
		nav = i;
	}
	if (i > 3) {
		char *tp = strdup(cmd);
		
		for (cp = strtok(tp, ","), Ac = 0; cp;
		     cp = strtok(NULL, ",")) {
			Av[Ac++] = cp;
		}
	} else {
		Ac = 1;
		Av[0] = cmd;
	}
	Ac += nargs;
	Nargs = nargs;
#ifdef MAIN
	printf("i = %d, nav = %d, Ac = %d\n", i, nav, Ac);
#endif	
	return Ac;
}


void
reapChildren()
{
	while (waitpid(-1, 0, WNOHANG) > 0)
		;
}

void
runChildAv(cmd, nargs, args, async)
	char *cmd;
	int nargs;
	char *args[];
	int async;
{
	int x, i, pid;

#ifdef MAIN
	printf("runChild('%s')\n", args ? args[0] : "");
#endif
	if (cmd)
		setChildAv(cmd, nargs);

	if (!Av)
		return;

	if (Nargs > 0) {
		for (i = 0, x = Ac - Nargs; x < Ac; ) {
			Av[x++] = args[i++];
		}
	}
	Av[Ac] = 0;

#ifdef MAIN
	for (i = 0; i < Ac; ++i)
		printf("\tAv[%d] = '%s'\n", i, Av[i]);
#endif
	for (pid = -1, i = 0; pid < 0 && i < 3; sleep(++i)) {
		switch(pid = vfork()) {
		case -1: syslog(LOG_ERR, "fork: %m"); break;
		case 0:
			execvp(Av[0], Av);
			syslog(LOG_ERR, "execvp(%s): %m", Av[0]);
			_exit(1);		/* NOT REACHED */
		default:
			syslog(LOG_INFO, "runChild%s[%d](%s)",
			       (async) ? "Async" : "",
			       pid, Av[0]);
			if (!async)
				(void) waitpid(pid, 0, 0);
			break;
		}
	}
	if (async) {
		reapChildren();
	}
}

void
runChild(cmd, arg, async)
	char *cmd;
	char *arg;
	int async;
{
	char *av[2];

	av[0] = arg;
	av[1] = 0;
	runChildAv(cmd, 1, av, async);
}
	
#ifdef MAIN
int
main(argc, argv)
	int argc;
	char *argv[];
{
	char *cmd;
	
	if (argc > 1)
		cmd = argv[1];
	else
		cmd = "test";
	runChild(cmd, "one", 0);
	runChild(0, "two", 1);
	runChild(0, "three", 1);
	exit(0);
}
#endif
